__version__ = '1.4'

def get(key, use_names=False):
    """
    Use this method to obtain topologies from the repository as dicts in node-link format.

    Example:

    .. code-block:: python

        import networkx as nx
        import topohub

        # Obtain topology dicts from JSON files stored in the repository
        topo = topohub.get('gabriel/25/0')
        topo = topohub.get('backbone/africa')
        topo = topohub.get('topozoo/Abilene')
        topo = topohub.get('sndlib/polska')

        # Create NetworkX graph from node-link dict
        g = nx.node_link_graph(topo)

        # Access graph parameters
        print(g.graph['name'])
        print(g.graph['demands'])
        print(g.graph['stats']['avg_degree'])

        # By default, nodes are referred by their integer IDs
        # You can obtain node names through the `name` attribute
        print(g.nodes[0]['name'])
        print(g.nodes[10]['name'])

        # Obtain link length in kilometers between node 0 and 10
        print(g.edges[0, 10]['dist'])
        # Obtain percentage utilization of the link between node 0 and 10 under ECMP routing in forward direction
        print(g.edges[0, 10]['ecmp_fwd']['uni'])

        # You can also load a topology using node names instead integer IDs as node identifiers
        # (this will not work for 'backbone' category topologies which have unnamed or duplicated name nodes)
        topo = topohub.get('sndlib/polska', use_names=True)
        g = nx.node_link_graph(topo)

        print(g.graph['demands'])
        print(g.edges['Gdansk', 'Warsaw']['dist'])
        print(g.edges['Gdansk', 'Warsaw']['ecmp_fwd']['uni'])

    """
    import importlib.resources
    import json

    import topohub.data

    try:
        topo = json.load((importlib.resources.files(topohub) / f'data/{key}.json').open())
        if use_names:
            id_to_name = {}
            name_to_id = {}
            for n in topo['nodes']:
                id_to_name[n['id']] = n['name']
                if n['name'] in name_to_id:
                    raise RuntimeError(f"Duplicate node name '{n['name']}'")
                else:
                    name_to_id[n['name']] = n['id']
            for n in topo['nodes']:
                n['id'] = id_to_name[n['id']]
            for l in topo['links']:
                l['source'] = id_to_name[l['source']]
                l['target'] = id_to_name[l['target']]
            topo['graph']['demands'] = {id_to_name[int(n)]: {id_to_name[int(k)]: v for k, v in dems.items()} for n, dems in topo['graph']['demands'].items()}
        return topo
    except IOError:
        raise KeyError



import mininet
import mininet.node
import mininet.topo

import topohub

class Demands(dict):
    """Demands dictionary"""
    def __missing__(self, key):
        return 0.0

    @property
    def maximum(self):
        """
        The maximum demand value in the dictionary.

        Returns
        -------
        float
            maximum demand value
        """
        try:
            return self.__maximum
        except AttributeError:
            try:
                self.__maximum = max(self.values())
            except ValueError:
                self.__maximum = 0.0
            return self.__maximum

    def get(self, k, default=0.0, sym=False, norm=False):
        """
        Return the demand value for (src, dst) node pair if specified in the dictionary, else default.

        Parameters
        ----------
        k : (str, str)
            (src, dst) node pair
        default : float, default 0.0
            default demand to return if not specified in the dictionary
        sym: bool, default False
            return demand for (dst, src) instead default if demand for (src, dst) not specified in dictionary
        norm: bool, default False
            normalize returned demands to [0.0 - 1.0]

        Returns
        -------
        float
            demand between pair of nodes
        """
        if k in self:
            val = self[k]
        else:
            src, dst = k
            if sym and (dst, src) in self:
                val = self[dst, src]
            else:
                val = default
        if norm:
            try:
                val /= self.maximum
            except ZeroDivisionError:
                val = 1.0
        return val

class JSONTopo(mininet.topo.Topo):
    demands = Demands()
    topo_json = {}

    def build(self, *args, **params):
        topo = self.topo_json

        if 'demands' in topo['graph']:
            demands = topo['graph']['demands']
            for src in demands:
                for dst in demands[src]:
                    self.demands[str(src), str(dst)] = demands[src][dst]

        for node in topo['nodes']:
            self.addSwitch(str(node['id']), **{k: v for k, v in node.items() if k not in ['id', 'name']})
        for link in topo['links']:
            self.addLink(str(link['source']), str(link['target']), **{k: v for k, v in link.items() if k not in ['source', 'target']})

        super(JSONTopo, self).build(**params)

def make_topo_class_from_json(name, topo):
    cls_name = name.title().replace('/', '_')
    cls = type(str(cls_name), (JSONTopo,), {'name': name, 'topo_json': topo})
    return cls

class TopoClsDict(dict):
    def __missing__(self, key):
        topo = topohub.get(key)
        cls = make_topo_class_from_json(key, topo)
        self[key] = cls
        return cls

class TopoNamedClsDict(dict):
    def __missing__(self, key):
        topo = topohub.get(key, use_names=True)
        cls = make_topo_class_from_json(key, topo)
        self[key] = cls
        return cls


TOPO_CLS = TopoClsDict()
"""
Use this dictionary to obtain topologies from the repository as Mininet Topo classes.

Example:

.. code-block:: python

    import mininet.net
    import topohub.mininet
    
    # Obtain Mininet Topo classes for topologies stored in the repository
    topo_cls = topohub.mininet.TOPO_CLS['gabriel/25/0']
    topo_cls = topohub.mininet.TOPO_CLS['backbone/africa']
    topo_cls = topohub.mininet.TOPO_CLS['topozoo/Abilene']
    topo_cls = topohub.mininet.TOPO_CLS['sndlib/polska']

    # Initialize Mininet Topo object
    topo = topo_cls()
    # Create Mininet Network using the selected topology
    net = mininet.net.Mininet(topo=topo)
    # Start the network and Mininet shell
    net.interact()
"""

TOPO_NAMED_CLS = TopoNamedClsDict()
"""
Use this dictionary to obtain topologies from the repository as Mininet Topo classes,
using node names instead integer IDs as node identifiers.

(this will not work for 'backbone' category topologies which have unnamed or duplicated name nodes)

Example:

.. code-block:: python

    import mininet.net
    import topohub.mininet

    # Obtain Mininet Topo classes for topologies stored in the repository
    topo_cls = topohub.mininet.TOPO_NAMED_CLS['gabriel/25/0']
    topo_cls = topohub.mininet.TOPO_NAMED_CLS['topozoo/Abilene']
    topo_cls = topohub.mininet.TOPO_NAMED_CLS['sndlib/polska']

    # Initialize Mininet Topo object
    topo = topo_cls()
    # Create Mininet Network using the selected topology
    net = mininet.net.Mininet(topo=topo)
    # Start the network and Mininet shell
    net.interact()
"""

class AutoHostTopo(mininet.topo.Topo):
    """Adds k hosts per switch"""

    def build(self, k=1, **params):
        super(AutoHostTopo, self).build(**params)

        for sw in self.switches():
            for i in range(k):
                h = self.addHost("h" + str(i) + "_" + sw)
                self.addLink(h, sw)

class AutoHostBridgeTopo(mininet.topo.Topo):
    """Adds b bridges, each with k hosts, per switch"""

    def build(self, b=1, k=1, **params):
        super(AutoHostBridgeTopo, self).build(**params)

        for sw in self.switches():
            for j in range(b):
                br = self.addSwitch("br" + str(j) + "_" + sw, cls=mininet.node.LinuxBridge, inNamespace=True)
                self.addLink(br, sw)
                for i in range(k):
                    h = self.addHost("h" + str(i) + "_" + br)
                    self.addLink(h, br, cls=mininet.link.TCLink)



import collections
import itertools
import math
import statistics
import time

def haversine(src, dst):
    """
    Calculate the great-circle distance between two points on a sphere given their longitudes and latitudes.

    Parameters
    ----------
    src : (float, float)
        coordinates (lon0, lat0)
    dst : (float, float)
        coordinates (lon1, lat1)

    Returns
    -------
    float
        distance in kilometers
    """

    (lon0, lat0) = src
    (lon1, lat1) = dst
    R = 6372.8
    d_lat = math.radians(lat1 - lat0)
    d_lon = math.radians(lon1 - lon0)
    lat0 = math.radians(lat0)
    lat1 = math.radians(lat1)
    a = math.sin(d_lat / 2) ** 2 + math.cos(lat0) * math.cos(lat1) * math.sin(d_lon / 2) ** 2
    return R * 2 * math.asin(math.sqrt(a))

def gini(list_of_values):
    """
    Calculate the Gini coefficient of a given list of values.

    Parameters
    ----------
    list_of_values : list[float]
        given list of values

    Returns
    -------
    float
        calculated Gini coefficient
    """

    sorted_list = sorted(list_of_values)
    height, area = 0, 0
    for value in sorted_list:
        height += value
        area += height - value / 2
    fair_area = height * len(list_of_values) / 2
    return (fair_area - area) / fair_area

def all_shortest_paths_all_targets(g, source, weight=None, limit=None):
    """
    Find all shortest paths from a source node to all other nodes.

    Parameters
    ----------
    g : networkx.Graph
        network graph
    source : object
        source node
    weight : str | function, default None
        name of edge attribute or a function returning edge weight
    limit : int, default None
        limit the number of shortest paths to find

    Returns
    -------
    dict[object, list], dict[object, int]
        paths and distances from the source node to all other nodes
    """

    import networkx as nx

    paths = {}

    if weight is not None:
        pred, dist = nx.dijkstra_predecessor_and_distance(g, source,
                                                          weight=weight)
    else:
        pred = nx.predecessor(g, source)
        dist = None

    if source not in g:
        raise nx.NodeNotFound('Source {} is not in G'.format(source))

    for target in g:

        if target not in pred:
            raise nx.NetworkXNoPath()

        paths_to_target = []

        stack = [[target, 0]]
        top = 0
        while top >= 0:
            node, i = stack[top]
            if node == source:
                paths_to_target.append([p for p, n in reversed(stack[:top + 1])])
                if limit and len(paths_to_target) == limit:
                    break
            if len(pred[node]) > i:
                top += 1
                if top == len(stack):
                    stack.append([pred[node][i], 0])
                else:
                    stack[top] = [pred[node][i], 0]
            else:
                stack[top - 1][1] += 1
                top -= 1

        paths[target] = paths_to_target

    return paths, dist

def all_shortest_nhops_all_targets(g, source, weight=None, limit=None):
    """
    Find all next hops from a source node to all other nodes.

    Parameters
    ----------
    g : networkx.Graph
        network graph
    source : object
        source node
    weight : str | function, default None
        name of edge attribute or a function returning edge weight
    limit : int, default None
        limit the number of shortest paths to find

    Returns
    -------
    dict[object, set], dict[object, int]
        next hops and distances from the source node to all other nodes
    """

    import networkx as nx

    nhops = {}

    if weight is not None:
        pred, dist = nx.dijkstra_predecessor_and_distance(g, source,
                                                          weight=weight)
    else:
        pred = nx.predecessor(g, source)
        dist = None

    if source not in g:
        raise nx.NodeNotFound('Source {} is not in G'.format(source))

    for target in g:

        if target not in pred:
            raise nx.NetworkXNoPath()

        nhops_to_target = set()

        stack = [[target, 0]]
        top = 0
        while top >= 0:
            node, i = stack[top]
            if node == source:
                nhops_to_target.add(stack[top - 1][0])
                if limit and len(nhops_to_target) == limit:
                    break
            if len(pred[node]) > i:
                top += 1
                if top == len(stack):
                    stack.append([pred[node][i], 0])
                else:
                    stack[top] = [pred[node][i], 0]
            else:
                stack[top - 1][1] += 1
                top -= 1

        nhops[target] = nhops_to_target

    return nhops, dist

def all_disjoint_paths(g, ff, node_filter=None):
    """
    Find all disjoint paths between all node pairs in a graph.

    Parameters
    ----------
    g : networkx.Graph
        network graph
    ff : function
        function for computing the maximum flow among a pair of nodes

    Returns
    -------
    dict[(object, object), list]
        dictionary of lists of disjoint paths between (src, dst) node pairs
    """

    import networkx as nx
    h = nx.algorithms.connectivity.build_auxiliary_edge_connectivity(g)
    r = nx.algorithms.flow.build_residual_network(h, 'capacity')
    result = {}
    for src, dst in itertools.permutations(g, 2):
        if node_filter is None or node_filter(g.nodes[src]) and node_filter(g.nodes[dst]):
            result[src, dst] = list(nx.edge_disjoint_paths(g, src, dst, auxiliary=h, residual=r, flow_func=ff))
    return result

def all_disjoint_paths_scipy(g, ff):
    """
    Find all disjoint paths between all node pairs in a graph using scipy.

    Parameters
    ----------
    g : networkx.Graph
        network graph
    ff : function
        function for computing the maximum flow among a pair of nodes

    Returns
    -------
    dict[(object, object), list]
        dictionary of lists of disjoint paths between (src, dst) node pairs
    """

    import networkx as nx
    import scipy.sparse
    mat = nx.convert_matrix.to_scipy_sparse_matrix(g)
    result = {}
    i_to_node = {}
    node_to_i = {}
    for n, node in enumerate(g):
        i_to_node[n] = node
        node_to_i[node] = n
    for s, t in itertools.permutations(g, 2):
        r = scipy.sparse.csgraph.maximum_flow(mat, node_to_i[s], node_to_i[t])
        R = nx.convert_matrix.from_scipy_sparse_matrix(r.residual, edge_attribute='flow', create_using=nx.DiGraph)
        # print(json.dumps(nx.node_link_data(R), indent=4, default=str))
        # Saturated edges in the residual network form the edge disjoint paths
        # between source and target
        cutset = [(i_to_node[int(u)], i_to_node[int(v)]) for u, v, d in R.edges(data=True)
                  if d['flow'] == 1]
        # This is equivalent of what flow.utils.build_flow_dict returns, but
        # only for the nodes with saturated edges and without reporting 0 flows.
        flow_dict = {n: {} for edge in cutset for n in edge}
        for u, v in cutset:
            flow_dict[u][v] = 1
        paths = []
        # Rebuild the edge disjoint paths from the flow dictionary.
        paths_found = 0
        for v in list(flow_dict[s]):
            path = [s]
            if v == t:
                path.append(v)
                paths.append(path)
                continue
            u = v
            while u != t:
                path.append(u)
                try:
                    u, _ = flow_dict[u].popitem()
                except KeyError:
                    break
            else:
                path.append(t)
                paths.append(path)
                paths_found += 1

        result[s, t] = paths

    return result

def shortest_disjoint_paths_slow(g):
    """
    Find all shortest disjoint paths between all node pairs in a graph.

    Parameters
    ----------
    g : networkx.Graph
        network graph

    Returns
    -------
    dict[(object, object), list]
        dictionary of lists of disjoint paths between (src, dst) node pairs
    """

    import networkx as nx
    result = {}
    for src, dst in itertools.permutations(g, 2):
        longest_paths = []
        for pp in itertools.permutations(nx.all_shortest_paths(g, src, dst, weight='cost')):
            paths = []
            used_edges = set()
            for path in pp:
                this_used_edges = set()
                for f, t in zip(path, path[1:]):
                    edge = (f, t)
                    if edge in used_edges:
                        break
                    else:
                        this_used_edges.add(edge)
                else:
                    paths.append(path)
                    used_edges |= this_used_edges
            if len(paths) > len(longest_paths):
                longest_paths = paths
        result[src, dst] = longest_paths
    return result

def shortest_disjoint_paths(g, ff, node_filter=None):
    """
    Find all shortest disjoint paths between all node pairs in a graph.

    Parameters
    ----------
    g : networkx.Graph
        network graph
    ff : function
        function for computing the maximum flow among a pair of nodes

    Returns
    -------
    dict[(object, object), list]
        dictionary of lists of disjoint paths between (src, dst) node pairs
    """

    import networkx as nx
    nhops = {}
    for n in g:
        nhops[n] = all_shortest_nhops_all_targets(g, n)[0]

    def dfs(src, dst, sg, ae):
        for nhop in nhops[src][dst]:
            edge = src, nhop
            if edge not in ae:
                sg.add_edge(src, nhop)
                ae.add(edge)
            if nhop != dst:
                dfs(nhop, dst, sg, ae)

    result = {}
    for src in g:
        for dst in g:
            if src != dst:
                if node_filter is None or node_filter(g.nodes[src]) and node_filter(g.nodes[dst]):
                    subgraph = nx.Graph()
                    added_edges = set()
                    dfs(src, dst, subgraph, added_edges)
                    result[src, dst] = list(nx.edge_disjoint_paths(subgraph, src, dst, flow_func=ff))
    return result

def save_topo_graph_pdf(g, filename=None, plot_aspect=1.0):
    """
    Generate and save topology graph as PDF file using matplotlib.

    Parameters
    ----------
    g : networkx.Graph
        network graph
    filename : str, default None
        filename
    plot_aspect : float, default 1.0
        plot aspect ratio
    """

    import networkx as nx
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt

    g = nx.relabel_nodes(g, lambda n: str(n))

    matplotlib.rcParams['pdf.use14corefonts'] = True
    matplotlib.rcParams['font.family'] = 'sans'

    pos = {}

    for node in g.nodes(data=True):
        pos[node[0]] = node[1]['pos']

    min0 = min(v[0] for v in pos.values())
    max0 = max(v[0] for v in pos.values())
    min1 = min(v[1] for v in pos.values())
    max1 = max(v[1] for v in pos.values())

    aspect = plot_aspect * (max0 - min0) / (max1 - min1)

    plt.figure(figsize=(2 * 3.4, 2 * 3.4 / aspect))
    nx.draw(g, pos=pos, node_size=100, alpha=0.25, node_color='k', edgelist=[])
    nx.draw_networkx_edges(g, pos=pos, node_size=100, alpha=0.25)
    nx.draw_networkx_labels(g, pos=pos)
    if not filename:
        filename = 'mininet/topo_lib/' + g.graph['name']
    plt.savefig(filename + '.pdf', metadata={'Creator': None, 'Producer': None, 'CreationDate': None})
    plt.close('all')

def minmax(pos):
    """
    Find min and max in node positions dictionary.

    Parameters
    ----------
    pos : dict
        node positions dictionary

    Returns
    -------
    (float, float, float, float)
        max_x, max_y, min_x, min_y
    """

    min0 = min(v[0] for v in pos.values())
    max0 = max(v[0] for v in pos.values())
    min1 = min(v[1] for v in pos.values())
    max1 = max(v[1] for v in pos.values())
    return max0, max1, min0, min1

def save_topo_graph_svg(g, filename=None, scale=None, background=None):
    """
    Generate and save topology graph as SVG file.

    Parameters
    ----------
    g : networkx.Graph
        network graph
    filename : str, default None
        filename
    scaling : bool, default True
        scale width and height of SVG to topology diameter
    """

    pos = {}
    names = {}
    for node in g.nodes(data=True):
        pos[node[0]] = node[1]['pos'][0], node[1]['pos'][1]
        names[node[0]] = node[1].get('name')

    max0, max1, min0, min1 = minmax(pos)

    if not scale:
        scale_factor = 1
    elif scale is True:
        scale_factor = max(abs(max0 - min0), abs(max1 - min1)) / 250
    else:
        scale_factor = scale

    stroke_width = scale_factor * 1
    circle_radius = scale_factor * 2
    font_size = scale_factor * 3

    with open(filename + '.svg', 'w') as f:
        f.write(f'<svg id="topo" viewBox="{(min0 - 0.05 * abs(max0 - min0)):.2f} {(-max1 - 0.05 * abs(max1 - min1)):.2f} {(1.10 * abs(max0 - min0)):.2f} {(1.10 * abs(max1 - min1)):.2f}" xmlns="http://www.w3.org/2000/svg">\n')
        f.write(f'<style>#topo path {{fill: none; stroke: grey; stroke-width: {stroke_width:.4f}px;}} #topo circle {{fill: lightblue; stroke: none; stroke-width: 0.3px; vector-effect: non-scaling-stroke; r: {circle_radius:.4f}px;}} #topo text {{fill: black; stroke: none; font-size: {font_size:.4f}px; font-family: sans-serif; text-anchor:middle; transform: translateY(0.3em);}} #topo path.country {{fill: lightgrey; stroke: black; stroke-width: 1px; vector-effect: non-scaling-stroke; visibility: visible;}} #topo path.selection {{fill: none; stroke: red; stroke-width: 1px; vector-effect: non-scaling-stroke; visibility: visible;}} #topo path.utilization {{visibility: hidden;}}</style>\n')

        if background:
            for item in background:
                f.write(item)

        for e in g.edges:
            x0, y0 = pos[e[0]]
            x1, y1 = pos[e[1]]
            f.write(f'<path data-id="{e[0]}-{e[1]}" d="M{x0:.2f},{-y0:.2f},{x1:.2f},{-y1:.2f}"/>\n')

        for n, (x, y) in pos.items():
            f.write(f'<circle data-id="{n}" cx="{x:.2f}" cy="{-y:.2f}" r="{circle_radius:.2f}px"/>\n')
            f.write(f'<text data-id="{n}" x="{x:.2f}" y="{-y:.2f}">{names[n] or n}</text>\n')

        f.write('</svg>\n')

def path_stats(g, node_filter=None):
    """
    Calculate statistics for paths between all node pairs.

    Parameters
    ----------
    g : networkx.Graph
        network graph

    Returns
    -------
    dict[(object, object), (int, int, float, int, float, float, float, object, object)]
        dictionary of paths statistics between (src, dst) node pairs in a format (adp_number, sdp_number, avg_adp_hops, avg_sdp_hops, avg_adp_length, avg_sdp_length, demand, src, dst)
    """

    import networkx as nx
    for ff in [
        nx.algorithms.flow.edmonds_karp,
        # nx.algorithms.flow.shortest_augmenting_path,
        # nx.algorithms.flow.preflow_push,
        # nx.algorithms.flow.dinitz,
        # nx.algorithms.flow.boykov_kolmogorov
    ]:
        ts = time.time()
        ap = all_disjoint_paths(g, ff, node_filter=node_filter)
        sp = shortest_disjoint_paths(g, ff, node_filter=node_filter)
        print(time.time() - ts)
    # sp_slow = shortest_disjoint_paths_slow(g)
    stats = {}
    for src, dst in ap:
        # assert len(sp[src, dst]) == len(sp_slow[src, dst])
        avg_ap_hops = sum(len(path) - 1 for path in ap[src, dst]) / float(len(ap[src, dst]))
        avg_sp_hops = sum(len(path) - 1 for path in sp[src, dst]) // len(sp[src, dst])
        avg_ap_length = 0
        for path in ap[src, dst]:
            for u, v in zip(path, path[1:]):
                avg_ap_length += g.edges[u, v]['dist']
        avg_ap_length /= float(len(ap[src, dst]))
        avg_sp_length = 0
        for path in sp[src, dst]:
            for u, v in zip(path, path[1:]):
                avg_sp_length += g.edges[u, v]['dist']
        avg_sp_length /= float(len(ap[src, dst]))
        if 'demands' in g.graph:
            try:
                demand = g.graph['demands'][src][dst]
            except KeyError:
                try:
                    demand = g.graph['demands'][dst][src]
                except KeyError:
                    demand = 0
        else:
            demand = 1
        stats[src, dst] = (len(ap[src, dst]), len(sp[src, dst]),
                           avg_ap_hops, avg_sp_hops,
                           avg_ap_length, avg_sp_length,
                           demand,
                           str(src), str(dst))
    for src, dst in stats:
        assert stats[src, dst][0:2] == stats[dst, src][0:2]
        # TODO: avg_ap_hops and avg_sp_hops are not symmetric (due to edmonds_karp not taking salts into account?)
    return stats

def path_stats_print(stats, filename=None):
    """
    Print statistics for paths between all node pairs in the CSV format.

    Parameters
    ----------
    stats : dict
        dictionary of paths statistics between (src, dst) node pairs
    filename : str, default None
        filename
    """

    table = sorted(stats.values(), reverse=True, key=lambda r: (r[0:2], sorted(r[-2:]), r[-1]))
    text = 'adp_number,sdp_number,avg_adp_hops,avg_sdp_hops,avg_adp_length,avg_sdp_length,demand,src,dst\n'
    text += '\n'.join(','.join(f'{v:.2f}' if isinstance(v, float) else str(v) for v in row) for row in table)
    if not filename:
        print(text)
    else:
        open(filename + '.csv', 'w').write(text)

def calculate_utilization(g, node_filter=None):
    """
    Calculate link utilizations for ECMP shortest path routing and save them as edges properties.

    Parameters
    ----------
    g : networkx.Graph
        network graph
    """

    nhops = {}
    for n in g:
        nhops[n] = all_shortest_nhops_all_targets(g, n)[0]

    modes = {
        'org': (((src, dst, g.graph['demands'][src][dst]) for src in g.graph['demands'] for dst in g.graph['demands'][src]) if 'demands' in g.graph else None),
        'uni': ((src, dst, 1) for src, dst in itertools.combinations(g, 2) if node_filter is None or node_filter(g.nodes[src]) and node_filter(g.nodes[dst])),
        'deg': ((src, dst, g.degree[src] * g.degree[dst]) for src, dst in itertools.combinations(g, 2) if node_filter is None or node_filter(g.nodes[src]) and node_filter(g.nodes[dst]))
    }

    def dfs(src, dst, dem, ut):
        dem /= len(nhops[src][dst])
        for nhop in nhops[src][dst]:
            ut[src, nhop] += dem
            if nhop != dst:
                dfs(nhop, dst, dem, ut)

    for mode, gen in modes.items():
        if gen:
            util = collections.defaultdict(float)
            for src, dst, dem in gen:
                dfs(src, dst, dem, util)
                dfs(dst, src, dem, util)
            if util:
                max_util = max(util.values())
                for src, dst in g.edges:
                    g.edges[src, dst].setdefault('ecmp_fwd', {})[mode] = util.get((src, dst), 0.0) / max_util * 100
                    g.edges[src, dst].setdefault('ecmp_bwd', {})[mode] = util.get((dst, src), 0.0) / max_util * 100

def topo_stats(g, ps=None):
    """
    Calculate topology properties statistics.

    Parameters
    ----------
    g : networkx.Graph
        network graph
    ps : dict, default None
        dictionary of paths statistics between (src, dst) node pairs

    Returns
    -------
    dict[str, int | float]
        dictionary topology properties
    """

    import networkx as nx

    min_link_length = min(e['dist'] for u, v, e in g.edges(data=True))
    avg_link_length = sum(e['dist'] for u, v, e in g.edges(data=True)) / float(len(g.edges))
    max_link_length = max(e['dist'] for u, v, e in g.edges(data=True))

    if isinstance(g, nx.MultiDiGraph):
        number_links = g.size() / 2
        degree = g.out_degree
    else:
        number_links = g.size()
        degree = g.degree
    degrees = [d for (_, d) in degree()]

    stats = {
        'nodes': len(g),
        'links': number_links,
        'demands': 0,
        'min_degree': min(degrees),
        'avg_degree': sum(degrees) / float(len(g)),
        'std_degree': statistics.pstdev(degrees),
        'max_degree': max(degrees),
        'gini': gini(degrees),
        'min_link_len': min_link_length,
        'avg_link_len': avg_link_length,
        'max_link_len': max_link_length,
        'diameter_len': nx.diameter(g, weight='dist'),
        'diameter_hops': nx.diameter(g)
    }

    if ps is None:
        dem_num = 0
        if 'demands' in g.graph:
            dem_num = sum(len(demands) for demands in g.graph['demands'].values())
        if dem_num == 0:
            dem_num = len(g) ** 2 - len(g)
    else:
        adp_sum, sdp_sum = 0, 0
        avg_adp_hops_sum, avg_sdp_hops_sum = 0.0, 0.0
        avg_adp_length_sum, avg_sdp_length_sum = 0.0, 0.0
        dem_num = 0
        dem_sum = 0.0
        max_adp_num = 0
        max_sdp_num = 0

        for pair, (adp_number, sdp_number, avg_adp_hops, avg_sdp_hops, avg_adp_length, avg_sdp_length, demand, src, dst) in ps.items():
            max_adp_num = max(max_adp_num, adp_number)
            max_sdp_num = max(max_sdp_num, sdp_number)
            dem_sum += demand

        for pair, (adp_number, sdp_number, avg_adp_hops, avg_sdp_hops, avg_adp_length, avg_sdp_length, demand, src, dst) in ps.items():
            if demand > 0 or dem_sum == 0:
                if dem_sum == 0:
                    demand = 1
                adp_sum += adp_number * demand
                sdp_sum += sdp_number * demand
                avg_adp_hops_sum += avg_adp_hops * demand
                avg_sdp_hops_sum += avg_sdp_hops * demand
                avg_adp_length_sum += avg_adp_length * demand
                avg_sdp_length_sum += avg_sdp_length * demand
                dem_num += 1

        if dem_sum == 0:
            dem_sum = dem_num

        stats['avg_sdp_num'] = sdp_sum / float(dem_sum)
        stats['max_sdp_num'] = max_sdp_num
        stats['avg_sdp_hops'] = avg_sdp_hops_sum / float(dem_sum)
        stats['avg_sdp_len'] = avg_sdp_length_sum / float(dem_sum)
        stats['avg_adp_num'] = adp_sum / float(dem_sum)
        stats['max_adp_num'] = max_adp_num
        stats['avg_adp_hops'] = avg_adp_hops_sum / float(dem_sum)
        stats['avg_adp_len'] = avg_adp_length_sum / float(dem_sum)

    stats['demands'] = dem_num
    return stats

def topo_stats_print(stats, name, filename=None):
    """
    Print topology statistics in a LaTeX format.

    Parameters
    ----------
    stats : dict
        dictionary topology properties
    name : str
        topology name
    filename : str, default None
        filename
    """

    just = 38
    text = \
        'Topology name'.ljust(just) + ' & %s' % name.replace('_', '\\_') + '\n\n' + \
        'Number of nodes'.ljust(just) + ' & %s' % stats['nodes'] + '\n' + \
        'Number of links'.ljust(just) + ' & %s' % stats['links'] + '\n' + \
        'Number of demands'.ljust(just) + ' & %s' % stats['demands'] + '\n' + \
        'Min. vertex degree'.ljust(just) + ' & %.2f' % stats['min_degree'] + '\n' + \
        'Avg. vertex degree'.ljust(just) + ' & %.2f' % stats['avg_degree'] + '\n' + \
        'Max. vertex degree'.ljust(just) + ' & %.2f' % stats['max_degree'] + '\n' + \
        'Min. link length'.ljust(just) + ' & %.2f' % stats['min_link_len'] + '\n' + \
        'Avg. link length'.ljust(just) + ' & %.2f' % stats['avg_link_len'] + '\n' + \
        'Max. link length'.ljust(just) + ' & %.2f' % stats['max_link_len'] + '\n'

    if 'avg_sdp_num' in stats:
        text += \
            'Avg. number of disjoint shortest paths'.ljust(just) + ' & %.2f' % stats['avg_sdp_num'] + '\n' + \
            'Max. number of disjoint shortest paths'.ljust(just) + ' & %.2f' % stats['max_sdp_num'] + '\n' + \
            'Avg. hops of disjoint shortest paths'.ljust(just) + ' & %.2f' % stats['avg_sdp_hops'] + '\n' + \
            'Avg. length of disjoint shortest paths'.ljust(just) + ' & %.2f' % stats['avg_sdp_len'] + '\n' + \
            'Avg. number of all disjoint paths'.ljust(just) + ' & %.2f' % stats['avg_adp_num'] + '\n' + \
            'Max. number of all disjoint paths'.ljust(just) + ' & %.2f' % stats['max_adp_num'] + '\n' + \
            'Avg. hops of all disjoint paths'.ljust(just) + ' & %.2f' % stats['avg_adp_hops'] + '\n' + \
            'Avg. length of all disjoint paths'.ljust(just) + ' & %.2f' % stats['avg_adp_len'] + '\n'

    if not filename:
        print(text)
    else:
        open(filename + '.tex', 'w').write(text)

def write_gml(g, path):
    """
    Writes a NetworkX graph to a GML format file.

    Parameters
    ----------
    g : networkx.Graph
        the input NetworkX graph
    path : str, default None
        the file path to write the GML file to

    """
    with open(path, 'w') as f:
        f.write('graph [\n')

        # Write graph attributes
        f.write(f'  name "{g.graph.get("name", "G")}"\n')
        f.write(f'  directed {int(g.is_directed())}\n')

        # Write stats
        stats = g.graph.get('stats')
        if stats:
            f.write('  stats [\n')
            for key, value in stats.items():
                f.write(f'    {key} {round(value, 2) if isinstance(value, float) else value}\n')
            f.write('  ]\n')

        # Write nodes
        for node, attr in g.nodes(data=True):
            f.write('  node [\n')
            f.write(f'    id {node}\n')
            f.write(f'    label "{attr["name"] if "name" in attr else node}"\n')
            if 'type' in attr:
                f.write(f'    type "{attr["type"]}"\n')
            if 'pos' in attr:
                f.write(f'    lon {round(attr["pos"][0], 2)}\n')
                f.write(f'    lat {round(attr["pos"][1], 2)}\n')
            f.write('  ]\n')

        # Write edges
        for source, target, attr in g.edges(data=True):
            f.write('  edge [\n')
            f.write(f'    source {source}\n')
            f.write(f'    target {target}\n')
            if 'type' in attr:
                f.write(f'    type "{attr["type"]}"\n')
            if 'dist' in attr:
                f.write(f'    dist {round(attr["dist"], 2)}\n')
            f.write('  ]\n')

        f.write(']')



#!/usr/bin/python3
import http.client as http
import json
import os
import random
import sys
import time

import networkx as nx

import topohub.graph

json.encoder.c_make_encoder = None
MAX_GABRIEL_NODES = 4096


class RoundingFloat(float):
    __repr__ = staticmethod(lambda x: format(x, '.2f'))


class TopoGenerator:

    @classmethod
    def generate_topo(cls, *args):
        return {}

    @classmethod
    def save_topo(cls, *args, **kwargs):
        g = nx.node_link_graph(cls.generate_topo(*args))
        if 'filename' in kwargs:
            filename = kwargs['filename']
        else:
            filename = f'mininet/topo_lib/{g.name}'
        os.makedirs(filename.rpartition('/')[0], exist_ok=True)
        ps = None
        if kwargs.get('with_plot'):
            topohub.graph.save_topo_graph_svg(g, filename, kwargs.get('scale'), kwargs.get('background'))
        if kwargs.get('with_utilization'):
            topohub.graph.calculate_utilization(g, node_filter=kwargs.get('node_filter'))
        if kwargs.get('with_path_stats'):
            ps = topohub.graph.path_stats(g, node_filter=kwargs.get('node_filter'))
            topohub.graph.path_stats_print(ps, filename)
        if kwargs.get('with_topo_stats'):
            ts = topohub.graph.topo_stats(g, ps)
            # graph.topo_stats_print(ts, g.graph['name'], filename)
            g.graph['stats'] = ts
        json.encoder.float = RoundingFloat
        json.dump(nx.node_link_data(g), open(f'{filename}.json', 'w'), indent=kwargs.get('indent', 0), default=lambda x: format(x, '.2f'))
        json.encoder.float = float
        topohub.graph.write_gml(g, f'{filename}.gml')

class SNDlibGenerator(TopoGenerator):
    """
    Generator of topologies from the SNDlib.

    Source of data: https://sndlib.put.poznan.pl

    Orlowski, S., Wessäly, R., Pióro, M. and Tomaszewski, A. (2010),
    SNDlib 1.0—Survivable Network Design Library. Networks, 55: 276-286.
    https://doi.org/10.1002/net.20371
    """

    @classmethod
    def download_topo(cls, name):

        con = http.HTTPSConnection("sndlib.put.poznan.pl", timeout=5)
        con.request('GET', "/download/sndlib-networks-native/%s.txt" % name)
        r = con.getresponse()
        data = r.read()
        return data

    @classmethod
    def generate_topo(cls, name):
        """
        Download topology specified by name from SNDlib and generate its JSON.

        Parameters
        ----------
        name : str
            topology name

        Returns
        -------
        dict
            topology graph in NetworkX node-link format
        """

        mode = None
        nodes = []
        links = []
        name_to_id = {}
        pos = {}
        demands = {}
        next_id = 0

        for line in cls.download_topo(name).splitlines():

            line = line.decode()

            if not mode:
                if line.startswith("NODES ("):
                    mode = 'nodes'
                elif line.startswith("LINKS ("):
                    mode = 'links'
                elif line.startswith("DEMANDS ("):
                    mode = 'demands'
                continue

            if mode == 'nodes':
                if line.startswith(")"):
                    mode = None
                    continue
                node = line.split()[0]
                lon, lat = line.split()[2:4]
                node_id = next_id
                next_id += 1
                name_to_id[node] = node_id
                pos[node] = (float(lon), float(lat))
                nodes.append({'id': node_id, 'name': node, 'pos': (float(lon), float(lat))})

            if mode == 'links':
                if line.startswith(")"):
                    mode = None
                    continue
                node0, node1 = line.split()[2:4]
                dist = topohub.graph.haversine(pos[node0], pos[node1])
                links.append({'source': name_to_id[node0], 'target': name_to_id[node1], 'dist': dist})

            if mode == 'demands':
                if line.startswith(")"):
                    mode = None
                    continue
                node0, node1 = line.split()[2:4]
                node0_id, node1_id = name_to_id[node0], name_to_id[node1]
                value = float(line.split()[6])
                if node0_id not in demands:
                    demands[node0_id] = {}
                demands[node0_id][node1_id] = value

        name = ''.join([c if c.isalnum() else '_' for c in name.title()])
        name = name.lower()

        return {'directed': False, 'multigraph': False, 'graph': {'name': name, 'demands': demands}, 'nodes': nodes, 'links': links}


class TopoZooGenerator(TopoGenerator):
    """
    Generator of topologies from the Internet Topology Zoo.

    Source of data: https://topology-zoo.org/

    S. Knight, H. X. Nguyen, N. Falkner, R. Bowden and M. Roughan,
    The Internet Topology Zoo. IEEE Journal on Selected Areas in Communications, vol. 29, no. 9, pp. 1765-1775.
    https://doi.org/10.1109/JSAC.2011.111002
    """

    @classmethod
    def download_topo(cls, name):

        con = http.HTTPSConnection("topology-zoo.org", timeout=5)
        con.request('GET', f"/files/{name}.gml")
        r = con.getresponse()
        data = r.read()
        return data

    @classmethod
    def generate_topo(cls, name):
        """
        Download topology specified by name from Topology Zoo and generate its JSON.

        Parameters
        ----------
        name : str
            topology name

        Returns
        -------
        dict
            topology graph in NetworkX node-link format
        """

        mode = None
        node_id, node, lon, lat, node0_id, node1_id = None, None, None, None, None, None
        nodes = []
        links = []
        pos = {}
        node_id_to_name = {}
        demands = {}

        for line in cls.download_topo(name).splitlines():

            line = line.decode()

            if not mode:
                if line.startswith("  node ["):
                    mode = 'node'
                    node_id, node, lon, lat = None, None, None, None
                elif line.startswith("  edge ["):
                    mode = 'edge'
                    node0_id, node1_id = None, None
                continue

            if mode == 'node':
                if line.startswith("  ]"):
                    if lon is not None and lat is not None:
                        node_id_to_name[node_id] = node
                        pos[node_id] = (float(lon), float(lat))
                        nodes.append({'id': node_id, 'name': node, 'pos': (float(lon), float(lat))})
                    mode = None
                elif line.startswith("    id "):
                    node_id = line.split()[-1]
                elif line.startswith("    label "):
                    node = line.split(maxsplit=1)[-1].strip('"')
                elif line.startswith("    Longitude "):
                    lon = line.split()[-1]
                elif line.startswith("    Latitude "):
                    lat = line.split()[-1]

            if mode == 'edge':
                if line.startswith("  ]"):
                    try:
                        dist = topohub.graph.haversine(pos[node0_id], pos[node1_id])
                        links.append({'source': node0_id, 'target': node1_id, 'dist': dist})
                    except KeyError:
                        pass
                    mode = None
                elif line.startswith("    source "):
                    node0_id = line.split()[-1]
                elif line.startswith("    target "):
                    node1_id = line.split()[-1]

        name = ''.join([c if c.isalnum() else '_' for c in name.title()])
        name = name.lower()

        if not nodes or not links:
            raise ValueError("Empty graph")

        return {'directed': False, 'multigraph': False, 'graph': {'name': name, 'demands': demands}, 'nodes': nodes, 'links': links}


class GabrielGenerator(TopoGenerator):
    """
    Generator of Gabriel graph synthetic topologies.

    E. K. Çetinkaya, M. J. F. Alenazi, Y. Cheng, A. M. Peck and J. P. G. Sterbenz,
    On the fitness of geographic graph generators for modelling physical level topologies.
    2013 5th International Congress on Ultra Modern Telecommunications and Control Systems and Workshops (ICUMT), Almaty, Kazakhstan, 2013, pp. 38-45.
    https://doi.org/10.1109/ICUMT.2013.6798402.

    K. Ruben Gabriel , Robert R. Sokal,
    A New Statistical Approach to Geographic Variation Analysis. Systematic Biology, Volume 18, Issue 3, September 1969, Pages 259–278.
    https://doi.org/10.2307/2412323
    """

    @classmethod
    def generate_topo(cls, nnodes, seed):
        """
        Generate Gabriel graph topology with given number of nodes.

        Parameters
        ----------
        nnodes : int
            number of nodes
        seed : int
            random seed

        Returns
        -------
        dict
            topology graph in NetworkX node-link format
        """

        assert nnodes <= MAX_GABRIEL_NODES

        nodes = []
        links = []
        pos = {}
        demands = {}

        random.seed(seed)

        scale = (nnodes * 10000) ** 0.5
        exclusion = 25.0  # prevent points from being too close
        print(nnodes, seed, scale, exclusion)

        def dist2(p, q):
            return (p[0] - q[0]) ** 2 + (p[1] - q[1]) ** 2

        while len(pos) < nnodes:
            p = (random.random() * scale, random.random() * scale)
            if pos:
                nn = min(dist2(p, q) ** 0.5 for q in pos.values())
                if nn < exclusion:
                    continue
            node_id = len(nodes)
            node = 'R' + str(node_id)
            pos[node_id] = p
            nodes.append({'id': node_id, 'name': node, 'pos': p})

        def neighbors(p, q):
            c = ((p[0] + q[0]) / 2, (p[1] + q[1]) / 2)
            dd = dist2(p, c)
            for r in pos.values():
                if r != p and r != q and dist2(r, c) < dd:
                    return False
            return True

        for p, (p_id, p_pos) in enumerate(pos.items()):
            for q, (q_id, q_pos) in enumerate(pos.items()):
                if p < q and neighbors(p_pos, q_pos):
                    dist = dist2(p_pos, q_pos) ** 0.5
                    links.append({'source': p_id, 'target': q_id, 'dist': dist})

        return {'directed': False, 'multigraph': False, 'graph': {'name': str(nnodes), 'demands': demands}, 'nodes': nodes, 'links': links}


class NumpyGabrielGenerator(TopoGenerator):
    """
    Generator of Gabriel graph synthetic topologies basen on NumPy.

    E. K. Çetinkaya, M. J. F. Alenazi, Y. Cheng, A. M. Peck and J. P. G. Sterbenz,
    On the fitness of geographic graph generators for modelling physical level topologies.
    2013 5th International Congress on Ultra Modern Telecommunications and Control Systems and Workshops (ICUMT), Almaty, Kazakhstan, 2013, pp. 38-45.
    https://doi.org/10.1109/ICUMT.2013.6798402.

    K. Ruben Gabriel , Robert R. Sokal,
    A New Statistical Approach to Geographic Variation Analysis. Systematic Biology, Volume 18, Issue 3, September 1969, Pages 259–278.
    https://doi.org/10.2307/2412323
    """

    @classmethod
    def generate_topo(cls, nnodes, seed):
        """
        Generate Gabriel graph topology with given number of nodes.

        Parameters
        ----------
        nnodes : int
            number of nodes
        seed : int
            random seed

        Returns
        -------
        dict
            topology graph in NetworkX node-link format
        """

        import numpy as np

        assert nnodes <= MAX_GABRIEL_NODES

        nodes = []
        links = []
        pos = {}
        demands = {}

        random.seed(seed)

        scale = (nnodes * 10000) ** 0.5
        exclusion = 25.0  # prevent points from being too close
        print(nnodes, seed, scale, exclusion)

        pos = np.empty((nnodes, 2))
        pos.fill(np.nan)

        def dist2(p, q):
            x = (p - q) ** 2
            return x[:, 0] + x[:, 1]

        n = 0
        exclusion_2 = exclusion ** 2
        while n < nnodes:
            p = (random.random() * scale, random.random() * scale)
            if nodes:
                if np.nanmin(dist2(p, pos)) < exclusion_2:
                    continue
            node_id = n
            node = f'R{node_id}'
            pos[node_id] = p
            nodes.append({'id': node_id, 'name': node, 'pos': p})
            n += 1

        def neighbors(p, q):
            c = (pos[p] + pos[q]) / 2
            dd = (pos[p] - c) ** 2
            dd = dd[0] + dd[1]
            # print(np.nanmin(dist2(c, pos)), dd)
            if dist2(pos, c).min() < dd:
                return False
            return True

        for p in range(nnodes):
            for q in range(nnodes):
                if p < q and neighbors(p, q):
                    dist = (pos[p] - pos[q]) ** 2
                    dist = (dist[0] + dist[1]) ** 0.5
                    links.append({'source': p, 'target': q, 'dist': dist})

        return {'directed': False, 'multigraph': False, 'graph': {'name': str(nnodes), 'demands': demands}, 'nodes': nodes, 'links': links}


class BackboneGenerator(TopoGenerator):
    """
    Generator of Gabriel graph synthetic topologies.

    E. K. Çetinkaya, M. J. F. Alenazi, Y. Cheng, A. M. Peck and J. P. G. Sterbenz,
    On the fitness of geographic graph generators for modelling physical level topologies.
    2013 5th International Congress on Ultra Modern Telecommunications and Control Systems and Workshops (ICUMT), Almaty, Kazakhstan, 2013, pp. 38-45.
    https://doi.org/10.1109/ICUMT.2013.6798402.

    K. Ruben Gabriel , Robert R. Sokal,
    A New Statistical Approach to Geographic Variation Analysis. Systematic Biology, Volume 18, Issue 3, September 1969, Pages 259–278.
    https://doi.org/10.2307/2412323
    """

    @classmethod
    def generate_topo(cls, name):
        """
        Generate Gabriel graph topology with given number of nodes.

        Parameters
        ----------
        nnodes : int
            number of nodes
        seed : int
            random seed

        Returns
        -------
        dict
            topology graph in NetworkX node-link format
        """
        import topohub.backbone

        j = json.load(open('external/backbone/graph.json'))

        nodes = {}
        names_set = set()
        links = []

        node_types = ['City', 'Seacable Landing Point']
        region = topohub.backbone.regions.get(name)
        if name.endswith('_nosc'):
            region = topohub.backbone.regions.get(name[:-5])
        else:
            node_types.append('Seacable Waypoint')

        for n in j['nodes']:
            if n['type'] in node_types:
                if not region or topohub.backbone.point_in_polygon(n['longitude'], n['latitude'], region):
                    if -180 < n['longitude'] < 180:
                        if n['type'] in ['City', 'Seacable Landing Point']:
                            names_set.add(n['name'])
                        else:
                            n['name'] = None
                        nodes[n['id']] = {'id': n['id'], 'pos': (n['longitude'], n['latitude']), 'type': n['type']}
                        if n['name']:
                            nodes[n['id']]['name'] = n['name']

        for e in j['edges']:
            if e['u'] in nodes and e['v'] in nodes:
                if not name.endswith('_nosc') or e['type'] != 'seacable':
                    if abs(nodes[e['u']]['pos'][0] - nodes[e['v']]['pos'][0]) < 180:
                        links.append({'source': e['u'], 'target': e['v'], 'dist': e['distance'], 'type': e['type']})

        g = nx.node_link_graph({'directed': False, 'multigraph': False, 'graph': {'name': str(name), 'demands': {}}, 'nodes': list(nodes.values()), 'links': links})
        g = topohub.backbone.remove_dead_ends(g)
        g = g.subgraph(max(nx.connected_components(g), key=len))

        return nx.node_link_data(g)


def main(topo_names):
    if topo_names[0] == 'gabriel':

        for nodes_number in range(25, 525, 25):
            start_time = time.time()
            for i in range(10):
                GabrielGenerator.save_topo(nodes_number, (i * MAX_GABRIEL_NODES) + nodes_number, filename=f'data/gabriel/{nodes_number}/{i}', with_plot=True, with_utilization=True, with_topo_stats=True, with_path_stats=True, scale=5)
            print(time.time() - start_time)

    elif topo_names[0] == 'sndlib':

        import topohub.backbone

        topo_names = {
            'abilene': {'include_countries': ['US']},
            'atlanta': None,
            'brain': {'include_countries': ['Germany']},
            'cost266': {'include_continents': ['EU']},
            'dfn-bwin': {'include_countries': ['Germany']},
            'dfn-gwin': {'include_countries': ['Germany']},
            'di-yuan': None,
            'france': None,
            'geant': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'germany50': {'include_countries': ['Germany']},
            'giul39': None,
            'india35': None,
            'janos-us': {'include_countries': ['US']},
            'janos-us-ca': {'include_countries': ['US', 'Canada']},
            'newyork': None,
            'nobel-eu': {'include_continents': ['EU']},
            'nobel-germany': {'include_countries': ['Germany']},
            'nobel-us': {'include_countries': ['US']},
            'norway': None,
            'pdh': {'include_countries': ['Germany']},
            'pioro40': None,
            'polska': {'include_countries': ['Poland']},
            'sun': None,
            'ta1': None,
            'ta2': None,
            'zib54': None
        }

        for topo_name in topo_names:
            if topo_names[topo_name]:
                background = topohub.backbone.generate_map(**topo_names[topo_name])
            else:
                background = None
            SNDlibGenerator.save_topo(topo_name, filename=f'data/sndlib/{topo_name}', with_plot=True, with_utilization=True, with_path_stats=True, with_topo_stats=True, background=background, scale=True)

    elif topo_names[0] == 'topozoo':

        import topohub.backbone

        topo_names = {
            'Aarnet': {'include_countries': ['Australia']},
            'Abilene': {'include_countries': ['US']},
            'Abvt': {'include_continents': ['EU'], 'include_countries': ['US', 'Japan']},
            'Aconet': {'include_countries': ['Austria']},
            'Agis': {'include_countries': ['US']},
            'Airtel': {'include_continents': ['EU', 'Asia'], 'include_countries': ['US']},
            # 'Ai3': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Amres': {'include_countries': ['Serbia']},
            'Ans': {'include_countries': ['United States of America']},
            'Arn': {'include_countries': ['Algeria']},
            'Arnes': {'include_countries': ['Slovenia']},
            'Arpanet196912': {'include_countries': ['US']},
            'Arpanet19706': {'include_countries': ['US']},
            'Arpanet19719': {'include_countries': ['US']},
            'Arpanet19723': {'include_countries': ['US']},
            'Arpanet19728': {'include_countries': ['US']},
            'AsnetAm': {'include_countries': ['US']},
            'Atmnet': {'include_countries': ['US']},
            'AttMpls': {'include_countries': ['US']},
            # 'Azrena': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'Bandcon': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Basnet': {'include_countries': ['Belarus']},
            'Bbnplanet': {'include_countries': ['US']},
            'Bellcanada': {'include_countries': ['US', 'Canada']},
            'Bellsouth': {'include_countries': ['US']},
            'Belnet2003': {'include_countries': ['Belgium']},
            'Belnet2004': {'include_countries': ['Belgium']},
            'Belnet2005': {'include_countries': ['Belgium']},
            'Belnet2006': {'include_countries': ['Belgium']},
            'Belnet2007': {'include_countries': ['Belgium']},
            'Belnet2008': {'include_countries': ['Belgium']},
            'Belnet2009': {'include_countries': ['Belgium']},
            'Belnet2010': {'include_countries': ['Belgium']},
            'BeyondTheNetwork': {'include_countries': ['US']},
            'Bics': {'include_continents': ['EU']},
            'Biznet': {'include_countries': ['Indonesia']},
            # 'Bren': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'BsonetEurope': {'include_continents': ['EU']},
            'BtAsiaPac': {'include_continents': ['Asia', 'Oceania']},
            'BtEurope': {'include_continents': ['EU']},
            # 'BtLatinAmerica': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'BtNorthAmerica': {'include_countries': ['US']},
            'Canerie': {'include_countries': ['US', 'Canada']},
            'Carnet': {'include_countries': ['Croatia']},
            'Cernet': {'include_countries': ['China']},
            'Cesnet1993': {'include_countries': ['Czechia']},
            'Cesnet1997': {'include_countries': ['Czechia']},
            'Cesnet1999': {'include_countries': ['Czechia']},
            'Cesnet2001': {'include_countries': ['Czechia']},
            'Cesnet200304': {'include_countries': ['Czechia']},
            'Cesnet200511': {'include_countries': ['Czechia']},
            'Cesnet200603': {'include_countries': ['Czechia']},
            'Cesnet200706': {'include_countries': ['Czechia']},
            'Cesnet201006': {'include_countries': ['Czechia']},
            'Chinanet': {'include_countries': ['China']},
            'Claranet': {'include_continents': ['EU']},
            # 'Cogentco': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'Colt': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'Columbus': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Compuserve': {'include_countries': ['US']},
            'CrlNetworkServices': {'include_countries': ['US']},
            # 'Cudi': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Cwix': {'include_countries': ['US']},
            'Cynet': {'include_countries': ['Cyprus']},
            'Darkstrand': {'include_countries': ['US']},
            'Dataxchange': {'include_countries': ['US']},
            # 'Deltacom': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'DeutscheTelekom': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Dfn': {'include_countries': ['Germany']},
            # 'DialtelecomCz': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Digex': {'include_countries': ['US']},
            # 'Easynet': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Eenet': {'include_countries': ['Estonia']},
            'EliBackbone': {'include_countries': ['US']},
            'Epoch': {'include_countries': ['US']},
            'Ernet': {'include_countries': ['India']},
            # 'Esnet': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'Eunetworks': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Evolink': {'include_countries': ['Bulgaria']},
            # 'Fatman': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Fccn': {'include_countries': ['Portugal']},
            'Forthnet': {'include_countries': ['Greece']},
            'Funet': {'include_countries': ['Finland']},
            'Gambia': {'include_countries': ['Gambia']},
            'Garr199901': {'include_countries': ['Italy']},
            'Garr199904': {'include_countries': ['Italy']},
            'Garr199905': {'include_countries': ['Italy']},
            'Garr200109': {'include_countries': ['Italy']},
            'Garr200112': {'include_countries': ['Italy']},
            'Garr200212': {'include_countries': ['Italy']},
            'Garr200404': {'include_countries': ['Italy']},
            'Garr200902': {'include_countries': ['Italy']},
            'Garr200908': {'include_countries': ['Italy']},
            'Garr200909': {'include_countries': ['Italy']},
            'Garr200912': {'include_countries': ['Italy']},
            'Garr201001': {'include_countries': ['Italy']},
            'Garr201003': {'include_countries': ['Italy']},
            'Garr201004': {'include_countries': ['Italy']},
            'Garr201005': {'include_countries': ['Italy']},
            'Garr201007': {'include_countries': ['Italy']},
            'Garr201008': {'include_countries': ['Italy']},
            'Garr201010': {'include_countries': ['Italy']},
            'Garr201012': {'include_countries': ['Italy']},
            'Garr201101': {'include_countries': ['Italy']},
            'Garr201102': {'include_countries': ['Italy']},
            'Garr201103': {'include_countries': ['Italy']},
            'Garr201104': {'include_countries': ['Italy']},
            'Garr201105': {'include_countries': ['Italy']},
            'Garr201107': {'include_countries': ['Italy']},
            'Garr201108': {'include_countries': ['Italy']},
            'Garr201109': {'include_countries': ['Italy']},
            'Garr201110': {'include_countries': ['Italy']},
            'Garr201111': {'include_countries': ['Italy']},
            'Garr201112': {'include_countries': ['Italy']},
            'Garr201201': {'include_countries': ['Italy']},
            'Gblnet': {'include_continents': ['EU']},
            'Geant2001': {'include_continents': ['EU'], 'include_countries': ['Israel', 'Cyprus']},
            'Geant2009': {'include_continents': ['EU'], 'include_countries': ['Israel', 'Turkey', 'Cyprus']},
            'Geant2010': {'include_continents': ['EU'], 'include_countries': ['Israel', 'Turkey', 'Cyprus']},
            'Geant2012': {'include_continents': ['EU'], 'include_countries': ['Israel', 'Turkey', 'Cyprus']},
            'Getnet': {'include_countries': ['US']},
            'Globalcenter': {'include_countries': ['US']},
            # 'Globenet': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Goodnet': {'include_countries': ['US']},
            'Grena': {'include_countries': ['Georgia']},
            'Gridnet': {'include_countries': ['US']},
            'Grnet': {'include_countries': ['Greece']},
            # 'GtsCe': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'GtsCzechRepublic': {'include_countries': ['Czechia']},
            'GtsHungary': {'include_countries': ['Hungary']},
            'GtsPoland': {'include_countries': ['Poland']},
            'GtsRomania': {'include_countries': ['Romania']},
            'GtsSlovakia': {'include_countries': ['Slovakia']},
            # 'Harnet': {'include_countries': ['Ireland']},
            'Heanet': {'include_countries': ['Ireland']},
            'HiberniaCanada': {'include_countries': ['Canada']},
            'HiberniaGlobal': {'include_continents': ['EU'], 'include_countries': ['US']},
            'HiberniaIreland': {'include_countries': ['Ireland']},
            'HiberniaNireland': {'include_countries': ['Ireland', 'United Kingdom']},
            'HiberniaUk': {'include_countries': ['United Kingdom']},
            'HiberniaUs': {'include_countries': ['US', 'Canada']},
            'Highwinds': {'include_continents': ['North America', 'South America', 'Europe'], 'exclude_countries': ['Russia']},
            'HostwayInternational': {'include_continents': ['all']},
            'HurricaneElectric': {'include_continents': ['all']},
            'Ibm': {'include_countries': ['US']},
            'Iij': {'include_countries': ['US', 'Japan']},
            'Iinet': {'include_countries': ['Australia']},
            'Ilan': {'include_countries': ['Israel']},
            'Integra': {'include_countries': ['US']},
            # 'Intellifiber': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Internetmci': {'include_countries': ['US']},
            'Internode': {'include_continents': ['all']},
            # 'Interoute': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'Intranetwork': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'Ion': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'IowaStatewideFiberMap': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Iris': {'include_countries': ['US']},
            'Istar': {'include_countries': ['Canada']},
            'Itnet': {'include_countries': ['Ireland']},
            'Janetbackbone': {'include_countries': ['United Kingdom']},
            # 'JanetExternal': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Janetlense': {'include_countries': ['United Kingdom']},
            'Jgn2Plus': {'include_countries': ['Japan']},
            'Karen': {'include_countries': ['New Zealand']},
            # 'Kdl': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'KentmanApr2007': {'include_countries': ['United Kingdom']},
            # 'KentmanAug2005': {'include_countries': ['United Kingdom']},
            'KentmanFeb2008': {'include_countries': ['United Kingdom']},
            # 'KentmanJan2011': {'include_countries': ['United Kingdom']},
            'KentmanJul2005': {'include_countries': ['United Kingdom']},
            'Kreonet': {'include_countries': ['South Korea']},
            # 'LambdaNet': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Latnet': {'include_countries': ['Latvia']},
            'Layer42': {'include_countries': ['US']},
            'Litnet': {'include_countries': ['Lithuania']},
            'Marnet': {'include_countries': ['North Macedonia']},
            'Marwan': {'include_countries': ['Morocco']},
            # 'Missouri': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Mren': {'include_countries': ['Montenegro']},
            'Myren': {'include_countries': ['Malaysia']},
            'Napnet': {'include_countries': ['US']},
            'Navigata': {'include_countries': ['Canada']},
            'Netrail': {'include_countries': ['US']},
            'NetworkUsa': {'include_countries': ['US']},
            'Nextgen': {'include_countries': ['Australia']},
            'Niif': {'include_countries': ['Hungary']},
            'Noel': {'include_countries': ['US', 'Canada']},
            'Nordu1989': {'include_continents': ['EU']},
            'Nordu1997': {'include_continents': ['EU']},
            'Nordu2005': {'include_continents': ['EU']},
            # 'Nordu2010': {'include_continents': ['EU']},
            # 'Nsfcnet': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Nsfnet': {'include_countries': ['US']},
            # 'Ntelos': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'Ntt': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'Oteglobe': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Oxford': {'include_countries': ['US']},
            'Pacificwave': {'include_countries': ['US']},
            'Packetexchange': {'include_continents': ['all']},
            # 'Padi': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Palmetto': {'include_countries': ['US']},
            'Peer1': {'include_continents': ['North America', 'Europe'], 'exclude_countries': ['Russia']},
            # 'Pern': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'PionierL1': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'PionierL3': {'include_countries': ['Poland']},
            'Psinet': {'include_countries': ['US']},
            'Quest': {'include_continents': ['all']},
            # 'RedBestel': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Rediris': {'include_countries': ['Spain']},
            'Renam': {'include_countries': ['Moldova']},
            'Renater1999': {'include_countries': ['France']},
            'Renater2001': {'include_countries': ['France']},
            'Renater2004': {'include_countries': ['France']},
            'Renater2006': {'include_countries': ['France']},
            'Renater2008': {'include_countries': ['France']},
            'Renater2010': {'include_countries': ['France']},
            'Restena': {'include_countries': ['Luxembourg']},
            # 'Reuna': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Rhnet': {'include_countries': ['Iceland']},
            'Rnp': {'include_countries': ['Brazil']},
            'Roedunet': {'include_countries': ['Romania']},
            # 'RoedunetFibre': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Sago': {'include_countries': ['US']},
            # 'Sanet': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Sanren': {'include_countries': ['South Africa']},
            'Savvis': {'include_countries': ['US']},
            # 'Shentel': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Sinet': {'include_countries': ['Japan']},
            # 'Singaren': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Spiralight': {'include_countries': ['US']},
            'Sprint': {'include_countries': ['US']},
            'Sunet': {'include_countries': ['Sweden']},
            'Surfnet': {'include_countries': ['Kingdom of the Netherlands']},
            # 'Switch': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'SwitchL3': {'include_countries': ['Switzerland']},
            # 'Syringa': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'TataNld': {'include_countries': ['India']},
            # 'Telcove': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Telecomserbia': {'include_countries': ['Serbia', 'Montenegro']},
            # 'Tinet': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'TLex': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'Tw': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'Twaren': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Ulaknet': {'include_countries': ['Turkey']},
            'UniC': {'include_countries': ['Denmark']},
            # 'Uninet': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Uninett2010': {'include_countries': ['Norway']},
            'Uninett2011': {'include_countries': ['Norway', 'Sweden']},
            'Uran': {'include_countries': ['Ukraine']},
            # 'UsCarrier': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            # 'UsSignal': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']},
            'Uunet': {'include_countries': ['US', 'Canada']},
            'Vinaren': {'include_continents': ['Asia', 'Oceania']},
            'VisionNet': {'include_countries': ['US']},
            'VtlWavenet2008': {'include_continents': ['EU']},
            'VtlWavenet2011': {'include_continents': ['EU']},
            'WideJpn': {'include_continents': ['all']},
            'Xeex': {'include_continents': ['all']},
            'Xspedius': {'include_countries': ['US']},
            'York': {'include_countries': ['United Kingdom']},
            # 'Zamren': {'include_continents': ['EU'], 'include_countries': ['US', 'Israel']}
        }

        for topo_name in topo_names:
            print(topo_name)
            if topo_names[topo_name]:
                background = topohub.backbone.generate_map(**topo_names[topo_name])
            else:
                background = None
            try:
                TopoZooGenerator.save_topo(topo_name, filename=f'data/topozoo/{topo_name}', with_plot=True, with_utilization=True, with_path_stats=True, with_topo_stats=True, background=background, scale=True)
            except nx.exception.NetworkXNoPath:
                pass
            except nx.NetworkXError:
                pass
            except ValueError as exc:
                if exc.args[0] == 'Empty graph':
                    pass
                else:
                    raise

    elif topo_names[0] == 'backbone':

        import topohub.backbone

        topo_names = {
            'africa': {'include_continents': ['Africa']},
            'africa_nosc': {'include_continents': ['Africa']},
            'americas': {'include_continents': ['North America', 'South America']},
            'americas_nosc': {'include_continents': ['North America', 'South America']},
            'atlantica': {'include_continents': ['North America', 'Europe'], 'include_countries': ['Turkey', 'Georgia', 'Cyprus'], 'exclude_countries': ['Russia']},
            'eastern': {'include_continents': ['Europe', 'Africa', 'Asia', 'Oceania']},
            'eastern_nosc': {'include_continents': ['Europe', 'Africa', 'Asia', 'Oceania']},
            'emea': {'include_continents': ['Europe', 'Africa', 'Asia']},
            'emea_nosc': {'include_continents': ['Europe', 'Africa', 'Asia']},
            'eurafrasia': {'include_continents': ['Europe', 'Africa', 'Asia']},
            'eurafrasia_nosc': {'include_continents': ['Europe', 'Africa', 'Asia']},
            'eurasia': {'include_continents': ['Europe', 'Asia']},
            'eurasia_nosc': {'include_continents': ['Europe', 'Asia']},
            'europe': {'include_continents': ['Europe'], 'include_countries': ['Turkey', 'Georgia', 'Cyprus'], 'exclude_countries': ['Russia']},
            'europe_nosc': {'include_continents': ['Europe'], 'include_countries': ['Turkey', 'Georgia', 'Cyprus'], 'exclude_countries': ['Russia']},
            'north_america': {'include_continents': ['North America']},
            'north_america_nosc': {'include_continents': ['North America']},
            'south_america': {'include_continents': ['South America']},
            'south_america_nosc': {'include_continents': ['South America']},
            'world': {'include_continents': ['all']}
        }

        for topo_name in topo_names:
            background = []
            if topo_names[topo_name]:
                background = topohub.backbone.generate_map(**topo_names[topo_name])
            # region = topohub.backbone.regions.get(topo_name)
            # if topo_name.endswith('_nosc'):
            #     region = topohub.backbone.regions.get(topo_name[:-5])
            # if region:
            #     path_data = topohub.backbone.polygon_to_path(region)
            #     background.append(f'<path class="selection" vector-effect="non-scaling-stroke" d="{path_data}"/>\n')
            BackboneGenerator.save_topo(topo_name, filename=f'data/backbone/{topo_name}', with_plot=True, with_utilization=True, with_path_stats=True, with_topo_stats=True, background=background, scale=0.1, node_filter=lambda n: n['type'] == 'City')


if __name__ == '__main__':
    main(sys.argv[1:])



from setuptools import setup
import topohub

URL = 'https://github.com/piotrjurkiewicz/topohub'

with open('README.md') as f:
    README = f.read()

setup(
    name='topohub',
    version=topohub.__version__,
    packages=['topohub'],
    package_dir={'topohub': 'topohub'},
    package_data={'topohub': ['data/gabriel/*/*.json', 'data/sndlib/*.json', 'data/topozoo/*.json', 'data/backbone/*.json']},
    url=URL,
    license='MIT',
    author='Piotr Jurkiewicz',
    author_email='piotr.jerzy.jurkiewicz@gmail.com',
    description='A repository of reference Gabriel graph and real-world topologies for networking research',
    long_description=README,
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Topic :: Internet",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: System :: Networking",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry"
    ]
)
